import subprocess
from abc import abstractmethod


class BaseDevice:
    def __init__(self, index: int = 0):
        self.index = index

    @abstractmethod
    def metrics(self):
        pass

    def __str__(self):
        return str(self.__dict__)

class BaseInfo:
    def __init__(
        self,
        type: str = "UNKNOWN",
        model: str = "UNKNOWN",
        vendor: str = "UNKNOWN",
        memory_total: int = 0,
    ):
        self.type = type
        self.model = model
        self.vendor = vendor
        self.memory_total = memory_total

    def __str__(self):
        return str(self.__dict__)

class BaseMetrics:
    def __init__(
        self,
        memory_used: int = 0,
        memory_process: int = 0,
        utilization: float = 0.0,
    ):
        self.memory_used = memory_used
        self.memory_process = memory_process
        self.utilization = utilization

    def __str__(self):
        return str(self.__dict__)


def _run(args) -> str:
    result = subprocess.run(
        args,
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
        text=True,
    )

    if result.returncode != 0 or result.stderr.strip() != "":
        raise RuntimeError(result.stderr)

    return result.stdout.strip()


import os

from .base import BaseDevice, BaseInfo, BaseMetrics, _run


class NvidiaGPU(BaseInfo):
    def __init__(self, features=[], *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.features = features


class NvidiaGPUMetrics(BaseMetrics):
    pass


class NvidiaDevice(BaseDevice):
    def __init__(self, cls, index: int = 0):
        super().__init__(index)
        self.gpu_id = self._get_gpu_id()

        try:
            args = [
                "nvidia-smi",
                f"--id={self.gpu_id}",
                "--query-gpu="
                "name,"
                "memory.total,"
                "pci.bus_id,"
                "pcie.link.gen.max,"
                "pcie.link.gen.current,"
                "driver_version",
                "--format=csv,noheader,nounits",
            ]

            result = _run(args=args)

            output = result.strip().split("\n")[0]
            model, total_memory, pci_bus_id, pcie_gen, pcie_width, driver_version = (output.split(", "))

            if model.lower().startswith("nvidia"):
                model = model[len("nvidia"):]

            compute_cap = (
                _run(["nvidia-smi", "--format=csv", "--query-gpu=compute_cap", "-i", f"{self.gpu_id}"])
                .removeprefix("compute_cap\n")
            )

            cls.type = "gpu"
            cls.model = model.strip().lower()
            cls.memory_total = int(total_memory) * 1024 * 1024  # bytes
            cls.vendor = "nvidia"
            cls.features = [compute_cap]
        except FileNotFoundError:
            raise FileNotFoundError()
        except Exception as e:
            raise e

    def _get_gpu_id(self):
        cudas = os.environ.get("CUDA_VISIBLE_DEVICES", "")
        cuda_list = cudas.split(",") if cudas else []
        if cuda_list and len(cuda_list) > self.index:
            return cuda_list[self.index]
        else:
            return str(self.index)

    def metrics(self):
        try:
            args = ["nvidia-smi", f"--id={self.gpu_id}", "--query-gpu=memory.used,utilization.gpu", "--format=csv,noheader,nounits", ]
            result = _run(args=args)

            output = result.split("\n")[0]
            used_memory, utilization = output.split(", ")

            return NvidiaGPUMetrics(
                memory_used=int(used_memory) * 1024 * 1024,  # bytes
                memory_process=0,  # Bytes, TODO, get this
                utilization=float(utilization),
            )

        except FileNotFoundError:
            raise FileNotFoundError(
                "The 'nvidia-smi' command was not found. Please ensure that the 'nvidia-utils' package is installed."
            )
        except Exception as e:
            raise e



import os
import platform
import subprocess

from .base import BaseDevice, BaseInfo, BaseMetrics, _run


class CPUInfo(BaseInfo):
    def __init__(self, features=[], *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.features = features


class CPUMetrics(BaseMetrics):
    pass


class CPUDevice(BaseDevice):
    def __init__(self, cls, index: int = 0):
        super().__init__(index)

        model = "Unknown Model"
        vendor = "Unknown vendor"
        flags = set()
        try:
            with open("/proc/cpuinfo", "r") as f:
                lines = f.readlines()
                for line in lines:
                    if line.startswith("flags"):
                        flags.update(line.strip().split(":")[1].split())
                    if line.startswith("model name"):
                        model = line.split(":")[1].strip()

                        if "amd" in model.lower():
                            if "epyc" in model.lower():
                                split = model.split(" ")
                                model = " ".join(split[1:3])
                            elif "ryzen" in model.lower():
                                split = model.split(" ")
                                model = " ".join(split[1:4])
                        elif "intel" in model.lower():
                            model = model.split(" ")[-1]

                    elif line.startswith("vendor_id"):
                        vendor = line.split(":")[1].strip()
        except FileNotFoundError:
            if platform.system() == "Darwin":
                model = (
                    _run(["sysctl", "-n", "machdep.cpu.brand_string"])
                    .replace("Apple", "")
                    .strip()
                )
                try:
                    vendor = (_run(["sysctl", "-n", "machdep.cpu.vendor"]))
                except subprocess.CalledProcessError:
                    vendor = "Apple"
            else:
                model = platform.processor()
                vendor = platform.uname().system

        if platform.system() == "Darwin":
            mem_total = int(_run(["sysctl", "-n", "hw.memsize"]))
            features = (
                _run(["sysctl -a | grep machdep.cpu.features"])
                .split(":")[1]
                .strip()
                .split()
            )

            flags = set(features)

        else:
            with open("/proc/meminfo", "r") as f:
                lines = f.readlines()
                mem_total = 0
                for line in lines:
                    if line.startswith("MemTotal:"):
                        mem_total = int(line.split()[1]) * 1024
                        break

        if "intel" in vendor.lower():
            vendor = "Intel"
        elif "amd" in vendor.lower():
            vendor = "AMD"

        cls.type = "cpu"
        cls.model = model.lower()
        cls.vendor = vendor.lower()
        cls.memory_total = mem_total  # Bytes
        self.memory_total = mem_total  # Bytes
        cls.features = sorted(set(f.lower() for f in flags))

    def _utilization(self):
        # check if is macOS
        if platform.system() == "Darwin":
            output = _run(["top", "-l", "1", "-stats", "cpu"])

            # CPU usage: 7.61% user, 15.23% sys, 77.15% idle
            for line in output.splitlines():
                if line.startswith("CPU usage"):
                    parts = line.split()
                    user_time = float(parts[2].strip("%"))
                    sys_time = float(parts[4].strip("%"))
                    idle_time = float(parts[6].strip("%"))
                    total_time = user_time + sys_time + idle_time
                    return total_time, idle_time
        else:
            with open("/proc/stat", "r") as f:
                lines = f.readlines()
                for line in lines:
                    if line.startswith("cpu "):
                        parts = line.split()
                        total_time = sum(int(part) for part in parts[1:])
                        idle_time = int(parts[4])
                        return total_time, idle_time

    def metrics(self):
        total_time_1, idle_time_1 = self._utilization()
        # read CPU status second time here, read too quickly will get inaccurate results
        total_time_2, idle_time_2 = self._utilization()

        total_diff = total_time_2 - total_time_1
        idle_diff = idle_time_2 - idle_time_1

        # total_diff might be 0
        if total_diff == 0:
            utilization = 0
        else:
            if platform.system() == "Darwin":
                utilization = idle_time_2 - idle_time_1
            else:
                utilization = (1 - (idle_diff / total_diff)) * 100

        if platform.system() == "Darwin":
            available_mem = _run(["vm_stat"])
            available_mem = available_mem.splitlines()

            free_pages = 0
            for line in available_mem:
                if "Pages free" in line:
                    free_pages = int(line.split(":")[1].strip().replace(".", ""))
                    break

            mem_free = free_pages * 16384
        else:
            with open("/proc/meminfo", "r") as f:
                lines = f.readlines()
                mem_free = 0
                for line in lines:
                    if line.startswith("MemAvailable:"):
                        mem_free = int(line.split()[1]) * 1024
                        break

        memory_used = self.memory_total - mem_free

        process_id = os.getpid()
        if platform.system() == "Darwin":
            result = _run(["ps", "-p", str(process_id), "-o", "rss="])
            memory_current_process = int(result) * 1024
        else:
            with open(f"/proc/{process_id}/status", "r") as f:
                lines = f.readlines()
                memory_current_process = 0
                for line in lines:
                    if line.startswith("VmRSS:"):
                        memory_current_process = int(line.split()[1]) * 1024
                        break

        return CPUMetrics(
            memory_used=memory_used,  # bytes
            memory_process=memory_current_process,  # bytes
            utilization=utilization,
        )



import platform
import re
import warnings

from .apple import AppleDevice
from .base import _run
from .cpu import CPUDevice
from .intel import IntelDevice
from .nvidia import NvidiaDevice

try:
    import torch

    HAS_TORCH = True
except BaseException:
    HAS_TORCH = False


class Device:
    def __init__(self, device):
        if HAS_TORCH and isinstance(device, torch.device):
            device_type = device.type.lower()
            device_index = device.index
        else:
            device_type = f"{device}".lower()
            device_index = 0
        if (
            device_type == "cuda"
            or device_type == "gpu"
            or device_type == "xpu"
            or re.match(r"(gpu|cuda|xpu):\d+", device_type)
        ):
            if platform.system() == "Darwin":
                if platform.machine() == 'x86_64':
                    raise Exception("Not supported for macOS on Intel chips.")

                self.device = AppleDevice(self, device_index)
            else:
                try:
                    result = _run(["lspci", "|", "grep", "-i", "vga\|3d\|display"])

                    output = result.lower()
                    if "intel" in output:
                        self.device = IntelDevice(self, device_index)
                    else:
                        self.device = NvidiaDevice(self, device_index)
                except BaseException:
                    self.device = NvidiaDevice(self, device_index)

        elif device_type == "cpu":
            self.device = CPUDevice(self, device_index)
        else:
            raise Exception(f"The device {device_type} is not supported")

    def info(self):
        warnings.warn(
            "info() method is deprecated and will be removed in next release.",
            DeprecationWarning,
            stacklevel=2
        )
        return self

    def memory_total(self):
        return self.memory_total

    def memory_used(self):
        return self.device.metrics().memory_used

    def utilization(self):
        return self.device.metrics().utilization

    def __str__(self):
        return str({k: v for k, v in self.__dict__.items() if k != 'device'})



from .base import BaseDevice, BaseInfo, BaseMetrics, _run


class AppleGPU(BaseInfo):
    pass  # TODO, add PCIE & DRIVER


class AppleGPUMetrics(BaseMetrics):
    pass


class AppleDevice(BaseDevice):
    def __init__(self, cls, index: int = 0):
        super().__init__(index)
        self.gpu_id = 0

        args = ["system_profiler", "SPDisplaysDataType"]

        result = _run(args=args)

        output = result.split("\n")
        model = ""
        vendor = ""
        for o in output:
            if "Chipset Model" in o:
                model = o.split(":")[1].replace("Apple", "").strip()
            if "Vendor" in o:
                vendor = o.split(":")[1].strip().split(" ")[0].strip()

        memory_total = int(_run(["sysctl", "-n", "hw.memsize"]))

        cls.type = "gpu"
        cls.model = model.lower()
        cls.memory_total = memory_total  # bytes
        cls.vendor = vendor.lower()

    def metrics(self):
        output = _run(["top", "-l", "1", "-stats", "cpu"])

        utilization = "0.0"
        for line in output.splitlines():
            if line.startswith("CPU usage"):
                parts = line.split()
                user_time = float(parts[2].strip("%"))
                sys_time = float(parts[4].strip("%"))
                utilization = user_time + sys_time

        total_memory = int(_run(['sysctl', 'hw.memsize']).split(':')[1].strip())
        free_memory = int(_run(['sysctl', 'vm.page_free_count']).split(':')[1].strip())
        page_size = int(_run(['sysctl', 'hw.pagesize']).split(':')[1].strip())

        used_memory = total_memory - (free_memory * page_size)

        return AppleGPUMetrics(
            memory_used=int(used_memory),  # bytes
            memory_process=0,  # Bytes, TODO, get this
            utilization=float(utilization),
        )



from setuptools import setup, find_packages

__version__ = "0.2.0"

setup(
    name="device-smi",
    version=__version__,
    author="ModelCloud",
    author_email="qubitium@modelcloud.ai",
    description="Retrieve gpu, cpu, and npu device info and properties from Linux/MacOS with zero package dependency.",
    long_description=open("README.md").read(),
    long_description_content_type="text/markdown",
    url="https://github.com/ModelCloud/Device-SMI/",
    packages=find_packages(),
    install_requires=[],
    platform=["linux"],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: OS Independent",
    ],
    python_requires=">=3",
)



import json

from .base import BaseDevice, BaseInfo, BaseMetrics, _run


class IntelGPU(BaseInfo):
    pass  # TODO, add PCIE & DRIVER


class IntelGPUMetrics(BaseMetrics):
    pass


class IntelDevice(BaseDevice):
    def __init__(self, cls, index: int = 0):
        super().__init__(index)
        self.gpu_id = index

        try:
            args = ["xpu-smi", "discovery", "-d", f"{self.gpu_id}", "-j"]

            result = _run(args=args)

            data = json.loads(result)

            model = data["device_name"]

            if model and model.lower().startswith("intel(r)"):
                model = model[8:].strip()
            vendor = data["vendor_name"]
            if vendor and vendor.lower().startswith("intel"):
                vendor = "Intel"
            total_memory = data["max_mem_alloc_size_byte"]

            cls.type = "gpu"
            cls.model = model.lower()
            cls.memory_total = int(total_memory)  # bytes
            cls.vendor = vendor.lower()

        except FileNotFoundError:
            raise FileNotFoundError("'xpu-smi' command not found. Please ensure it is installed")
        except Exception as e:
            raise e

    def metrics(self):
        try:
            args = [
                "xpu-smi", "dump",
                "-d", f"{self.gpu_id}",
                "-m", "0,18",
                "-n", "1"
            ]
            result = _run(args=args)

            # xpu-smi dump -d 0 -m 0,1,2 -i 1 -n 5
            # Timestamp, DeviceId, GPU Utilization (%), GPU Power (W), GPU Frequency (MHz)
            # 06:14:46.000,    0, 0.00, 14.61,    0

            output = result.split("\n")[-1]
            memory_used = output.split(",")[-1].strip()
            utilization = output.split(",")[-2].strip()
            if utilization.lower() == "n/a":
                utilization = "0.0"

            return IntelGPUMetrics(
                memory_used=int(float(memory_used) * 1024 * 1024),  # bytes
                memory_process=0,
                utilization=float(utilization),
            )
        except FileNotFoundError:
            raise FileNotFoundError("'xpu-smi' command not found. Please ensure it is installed")
        except Exception as e:
            raise e
